
/*
bosh main file

$Id: bosh.c,v 1.85 2008/12/19 16:51:45 alexsisson Exp $

(C) Copyright 2004-2008 Alex Sisson (alexsisson@gmail.com)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <ncurses.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/wait.h>
#include <errno.h>
#include <time.h>

#include "bosh.h"
#include "list.h"
#include "rc.h"
#include "system.h"
#include "search.h"
#include "ncursesex.h"
#include "misc.h"
#include "action.h"

/* defines */
#define HEADER 2
#define FOOTER 2
#define FRAME  (HEADER+FOOTER)

#define lof(a,b) (a<b?a:b) /* lower of */
#define hof(a,b) (a>b?a:b) /* higher of */
#define errstr   (strerror(errno))

/* globals */
char   *REPLY = NULL;      /* input return */
int    ULINES;             /* usable lines */
char   *conf = NULL;
char   *confpath = NULL;
BOSH   *bosh;              /* pointer to current BOSH struct */
LIST   *boshlist;          /* linked list for BOSH structs */
int    boshlistpos = 0;
MEVENT *mouse = NULL;      /* ncurses mouse struct - unused at the moment */
char   *search = NULL;     /* previous search */
time_t autorefreshtime;
char   *shell = "/bin/sh";
char   *bosherr = NULL;
char   *dollarbosh = NULL;
char   *dollarboshparam = NULL;
char   dollarbosherrfile[16];

#ifdef DEBUG
int debug_leavescript = 1;
#endif

#ifdef LOG
FILE *bosh_log = NULL;
#endif

/* functions */
void finish(int s) {
  if(s==SIGSEGV) {
    bosherr = "received SIGSEGV :-(";
  } else {
    erase();
    refresh();
    endwin();
  }
#ifdef LOG
  fprintf(bosh_log,"sig: %d - exiting (%s)\n",s,bosherr?bosherr:"");
#endif
  if(bosherr)
    fprintf(stderr,"%s: %s\n",PACKAGE,bosherr);
  exit(s);
}

void ctrlc(int s) {
  ungetch(3);
}

int mainloop(int mode);
int readstr(char *prompt);
int readstrp(char *prompt, char *previous);

int keyup(int n);
int keydown(int n);

BOSH *bosh_new() {
  BOSH *r;
  r = malloc(sizeof(BOSH));
  if(r) {
    memset(r,0,sizeof(BOSH));
    r->exit = -1;
    r->fd = -1;
    r->fderr = -1;
    r->width = BOSH_WIDTH_DEFAULT;
    r->cursorsize = 1;
    r->cursormovement = 1;
    r->buf = malloc(sizeof(LIST));
    list_init(r->buf);
    r->searchwrap = 1;
  }
  return r;
}

int bosh_free(BOSH *b) {
  int i;
  free(b->command);
  free(b->pipe);
  free(b->preaction);
  if(b->multilineseperator)
    if(strlen(b->multilineseperator))
      free(b->multilineseperator);
  for(i=0;i<37;i++)
    free(b->action[i].command);
  close(b->fd);
  close(b->fderr);
  if(b->childpid)
    kill(b->childpid,9);
  free(b);
  return 0;
}

/* main */
int main(int argc, char **argv) {

#ifdef LOG
  bosh_log = fopen("bosh.log","w");
#endif

  /* init bosh */
  boshlist = malloc(sizeof(LIST));
  list_init(boshlist);
  bosh = bosh_new();
  list_add(boshlist,bosh);

  bosh_parseargs(&argc,&argv);

  if(bosh->command) {
    bosh_open(bosh);
    if(bosh->width<strlen(bosh->command))
      bosh->width = strlen(bosh->command);
  }

  strcpy(dollarbosherrfile,"/tmp/boshXXXXXX");
  close(mkstemp(dollarbosherrfile));

  if(!isatty(0)) {
    /* piped input */
    bosh->fd = dup(0);
    dup2(1,0);
    bosh->command = NULL;
  } else if(!bosh->command) {
    fprintf(stderr,"%s: no input\n",PACKAGE);
    return 1;
  }

  /* set up signal handlers */
  signal(SIGTERM,finish);
  signal(SIGHUP,finish);
  signal(SIGSEGV,finish);
  signal(SIGINT,ctrlc);
  signal(SIGPIPE,SIG_IGN);

  /* init ncurses */
  initscr();
  keypad(stdscr,TRUE);
  cbreak();
  noecho();
  halfdelay(2);
  if(mouse)
    mousemask(BUTTON1_PRESSED,NULL);

  if(bosh->autorefresh>0)
    autorefreshtime = time(NULL);

  keyup(0);

  /* start main loop */
  mainloop(0);
  return 0;
}

#define FD_ISVALID(fd) (fd>=0)
#define FD_ISVALIDANDSET(fd,fdset) (fd>=0?FD_ISSET(fd,fdset):0)
#define FD_SETIFVALID(fd,fdset,n) if(fd>=0) { FD_SET(fd,fdset); n = hof(n,fd); }
#define debug(s) list_adddup(bosh->buf,s,strlen(s)+1)

int mainloop(int mode) {
  int key,i;
  char c;
  static char readbuf[1024];
  int readbuflen = 0;
  fd_set fdsetr;
  fd_set fdsetw;
  int n = 0;
  int loop;
  int needredraw;
  FILE *err;

  while(1) {
    loop = 1;
    needredraw = 0;
/*
    if the cursor is currently above the buffer length, it suggests that the input has
    been refreshed, so we wait a little while so that so data can arrive. This way, the
    refresh code won't just "correct" the problem, and lose cursor position.
*/
    if(bosh->line>list_items(bosh->buf))
      usleep(BOSH_INPUTWAIT_USLEEP);
    if(bosh->autorefresh>0) {
      if(time(NULL)>=autorefreshtime+bosh->autorefresh) {
        bosh_open(bosh);
        autorefreshtime = time(NULL);
      }
    }

    while(loop) {
      struct timeval tv = {0,0};
      loop = 0;
      FD_ZERO(&fdsetr);
      FD_ZERO(&fdsetw);
      FD_SETIFVALID(bosh->fd,     &fdsetr,n);
      FD_SETIFVALID(bosh->fderr,  &fdsetr,n);
      if(select(n+1,&fdsetr,&fdsetw,0,&tv)>0) {
        n = -1;
        if(FD_ISVALIDANDSET(bosh->fd,&fdsetr))
          n = bosh->fd;
        else if(FD_ISVALIDANDSET(bosh->fderr,&fdsetr))
          n = bosh->fderr;
        if(FD_ISVALID(n)) {
          if(read(n,&c,1)>0) {
            if(c=='\n') {
              readbuf[readbuflen] = 0;
              list_adddup(bosh->buf,readbuf,readbuflen+1);
              if(bosh->width<readbuflen)
                bosh->width = readbuflen;
              readbuflen = 0;
              needredraw = 1;
            } else if(readbuflen<sizeof(readbuf)) {
              readbuf[readbuflen] = c;
              readbuflen++;
            }
            loop = 1;
          } else {
            close(bosh->fd);
            close(bosh->fderr);
            bosh->fd = -1;
            bosh->fderr = -1;
            if(bosh->childpid) {
              /* need to reap child */
              waitpid(bosh->childpid,&n,0);
              bosh->exit = WIFEXITED(n) ? WEXITSTATUS(n) : -1;
              bosh->childpid = 0;
            }
#ifdef DEBUG
            if(!debug_leavescript)
#endif
            if(bosh->script)
              unlink(bosh->script);

            err = fopen(dollarbosherrfile,"r");
            if(err) {
              bosherr = malloc(256);
              fgets(bosherr,256,err);
              fclose(err);
              unlink(dollarbosherrfile);
              if(strlen(bosherr)) {
                char *p = strchr(bosherr,'\n');
                if(p)
                  *p = 0;
                finish(bosh->exit);
              }
              free(bosherr);
              bosherr = NULL;
            }
          }
        }
      }
    }

    key = getch();
    // This breaks KEY_RESIZE handling ??????????????
    // if(key==ERR && !needredraw)
    //   continue;

#ifdef LOG
    if(key!=ERR)
      fprintf(bosh_log,"key: %03d (%d)\n",key,mode);
#endif

    switch(mode) {

      case BOSH_MODE_NORMAL:
        switch(key) {
          case KEY_UP:    keyup(0);                    break;
          case KEY_DOWN:  keydown(0);                  break;
          case KEY_PPAGE: for(i=0;i<8;i++) keyup(0);   break;
          case KEY_NPAGE: for(i=0;i<8;i++) keydown(0); break;

          case KEY_LEFT:
            if(boshlistpos) {
              boshlistpos--;
              bosh = list_get(boshlist,boshlistpos);
              if(bosh->refresh)
                bosh_open(bosh);
            }
            break;
          case KEY_RIGHT:
            if(boshlistpos+1<list_items(boshlist)) {
              boshlistpos++;
              bosh = list_get(boshlist,boshlistpos);
              if(bosh->refresh)
                bosh_open(bosh);
            }
            break;
          case KEY_TAB:
            break;
          case KEY_SPACE:
            break;

          case KEY_CTRLA:
            readstr("action");
            if(REPLY) {
//              bosh->action[
            }
            break;

/*
old save code
            readstr("file");
            if(REPLY) {
              rc_write(REPLY,bosh);
              free(conf);
              conf = strdup(REPLY);
            }
            break;
*/

          /* jump to line */
          case KEY_CTRLJ:
            readstr("line");
            if(REPLY) {
              int line = strtol(REPLY,0,10);
              if(line>0) {
                line--;
                if(line<bosh->line)
                  while(line<bosh->line)
                    keyup(1);
                else
                  while(line>bosh->line && keydown(1));
              }
            }
            break;

          /* delete item from list */
          case KEY_CTRLK:
          case KEY_DC:
            list_del(bosh->buf,bosh->line);
            if(bosh->line>=list_items(bosh->buf))
              ungetch(KEY_UP);
            break;

          /* run new command */
          case KEY_CTRLO:
            readstr("command");
            if(REPLY) {
              free(bosh->command);
              bosh->command = strdup(REPLY);
              ungetch(KEY_CTRLR);
            }
            break;

          /* pipe buffer through command */
          case KEY_CTRLP:
          case '|':
            readstrp("pipe",bosh->pipe);
            if(REPLY) {
              if(strlen(REPLY)) {
                free(bosh->pipe);
                bosh->pipe = strdup(REPLY);
              }
              if(bosh->pipe) {
                if(bosh_pipe(bosh,bosh->pipe)==0) {
                  char *p;
                  if(bosh->command) {
                    p = malloc(strlen(bosh->command)+strlen(bosh->pipe) + 4);
                    sprintf(p,"%s | %s",bosh->command,bosh->pipe);
                    free(bosh->command);
                  } else {
                    p = malloc(strlen(bosh->pipe) + 3);
                    sprintf(p,"| %s",bosh->pipe);
                  }
                  bosh->command = p;
                }
              }
            }
            break;

          /* refresh */
          case KEY_CTRLR:
          case KEY_F(5):
            if(conf)
              bosh_open(bosh);
            break;

          /* insert blank line into list */
          case KEY_CTRLU:
          case KEY_IC:
            list_insdup(bosh->buf,bosh->line,"\0",1);
            break;

          /* search */
          case KEY_F(3):
          case KEY_CTRLW:
          case '/':
            readstrp("search",search);
            if(!REPLY)
              break;
            if(strlen(REPLY)) {
              bosh_regex_init(REPLY);
              free(search);
              search = strdup(REPLY);
#ifdef LOG
              fprintf(bosh_log,"search: new search \"%s\"\n",search);
#endif
            }
          case KEY_F(4):
          case KEY_CTRLN:
            if(search) {
              char *p;
              int startline = bosh->line;
#ifdef LOG
              fprintf(bosh_log,"search: starting search \"%s\"\n",search);
#endif
              keydown(1);
              while(1) {
                if(startline==bosh->line)
                  break;
                p = list_get(bosh->buf,bosh->line);
                if(!p)
                  break;
#ifdef LOG
                fprintf(bosh_log,"search: current line %d: \"%s\"\n",bosh->line,p);
#endif
                if(!bosh_regex_try(p)) /* match */
                  break;
                if(bosh->line+1>=list_items(bosh->buf)) {
                  /* end of list */
                  if(bosh->searchwrap) {
                    bosh->line = bosh->offset = bosh->cursor = 0;
                    continue;
                  }
                  else
                    break;
                }
                if(!keydown(1))
                  break;
              }
            }
            break;

          case KEY_CTRLV:
            action_run(BOSH_ACTION_SPECIAL_SHOWCONF,NULL);
            break;

          case KEY_CTRLX:
            finish(0);

          case KEY_MOUSE:
            getmouse(mouse);
            //if(mouse->x > 2-2 && mouse->x < 2+2)
            //ungetch(keyyuowant)
            break;

          /* reload config */
          case KEY_F(6):
//            if(confpath) {
//              rc_read(confpath,bosh);
//              bosh_open(bosh);
//            }
            break;

          /* redraw */
          case KEY_F(12):
          case KEY_CTRLL:
            erase();
          case KEY_RESIZE:
            refresh();
            break;

          /* action */
          default:
            key = isdigit(key) ? key - 48 : isupper(key) ? key - 55 : islower(key) ? key - 87 : key=='\n' ? 36 : -1;
            if(key>=0)
              if(bosh->action[key].command) {
                if(bosh->action[key].prompt) {
                  readstr(bosh->action[key].prompt);
                  if(REPLY)
                    action_run(key,REPLY);
                } else
                  action_run(key,NULL);
              }
            break;
        }
        bosh_redraw();
        break;

      case BOSH_MODE_READSTR:
        /* readstr */
        switch(key) {
          case KEY_CTRLX:
            finish(0);
          case KEY_CTRLC:
            return -1;
          case KEY_CTRLH:
          case KEY_BKSPC:
          case KEY_BACKSPACE:
            if(strlen(REPLY)) {
              REPLY[strlen(REPLY)-1] = 0;
              mvaddch(stdscr->_cury,stdscr->_curx-1,' ');
              move(stdscr->_cury,stdscr->_curx-1);
            }
            break;
          case '\n':
            return 0;
          default:
            if(isprint(key)) {
              REPLY[strlen(REPLY)] = key;
              addch(key);
              refresh();
            }
            break;
        }
        break;
    }
  }
}

/* redraw */
int bosh_redraw() {
  int i;
  char *p;
  ULINES = LINES - FRAME;
  erase();
//  Ensure cursor is in buffer range. Commented out because it was being called on a refresh
//  while(bosh->offset+bosh->cursor>list_items(bosh->buf)+1)
//    keyup();
  attron(A_REVERSE);
  mvaddstr(0,0,strmak(COLS,' '));
  if(conf)
    mvprintw(0,0,"%s",conf);
  mvprintw(0,COLS-(strlen(PACKAGE)+strlen(VERSION)+1),"%s %s",PACKAGE,VERSION);
  attroff(A_REVERSE);
  list_start(bosh->buf,bosh->offset);  
  for(i=0;i<ULINES;i++) {
    p = list_next(bosh->buf);
    if(p) {
      if(i>=bosh->cursor && i<bosh->cursor+bosh->cursorsize)
        attron(A_REVERSE);
      mvaddstr(HEADER+i,0,strmak(lof(COLS,bosh->width),' '));
      mvaddstr(HEADER+i,0,p);
      //  this makes bosh crash :-( - linked list could be dodgy:   addstr(strmak(lof(COLS,bosh->width)-strlen(p),' '));
      if(i>=bosh->cursor && i<bosh->cursor+bosh->cursorsize)
        attroff(A_REVERSE);
    } else {
      mvaddstr(HEADER+i,0,strmak(lof(COLS,bosh->width),' '));
    }
  }

  /* bottom lines */
  attron(A_REVERSE);
//  mvaddstr(LINES-2,0,strmak(lof(COLS,bosh->width),' '));
//  mvprintw(LINES-2,0,"%s",bosh->command?bosh->command:"[stdin]");
  mvaddstr(LINES-1,0,strmak(lof(COLS,bosh->width),' '));
  if(bosh->cursorsize>1)
    mvprintw(LINES-1,0,"%d-%d/%d ",bosh->line+1,lof(bosh->line+1+bosh->cursorsize-1,list_items(bosh->buf)),list_items(bosh->buf));
  else
    mvprintw(LINES-1,0,"%d/%d ",bosh->line,list_items(bosh->buf));
  printw(" exit=");
  if(bosh->exit>=0)
    printw("%d",bosh->exit);
  else
    printw(" ");
  printw(" %s ",bosh->refresh?"R":" ");
  if(bosh->autorefresh)
    printw("%d",bosh->autorefresh-(time(NULL)-autorefreshtime));
  attroff(A_REVERSE);
  move(LINES-1,COLS-1);

  refresh();
  return 0;
}

int readstr(char *prompt) {
  int n;
  free(REPLY);
  REPLY = malloc(128);
  if(!REPLY)
    return -1;
  memset(REPLY,0,128);
  attron(A_REVERSE);
  mvaddstr(LINES-1,0,strmak(bosh->width,' '));
  mvprintw(LINES-1,0,"%s: ",prompt);
  n = mainloop(1);
  attroff(A_REVERSE);
  if(!n)
    return 0;
  free(REPLY);
  REPLY = NULL;
  return -1;
}

int readstrp(char *prompt, char *previous) {
  char *s;
  int r;
  if(previous) {
    s = malloc(strlen(prompt)+strlen(previous)+4);
    sprintf(s,"%s [%s]",prompt,previous);
    r = readstr(s);
    free(s);
  } else {
    r = readstr(prompt);
  }
  return r;
}

int keyup(int n) {
  if(!n)
    n = bosh->cursormovement;
  if(bosh->cursor >= n)
    bosh->cursor -= n;
  else if(bosh->offset >= n)
    bosh->offset -= n;
  if(bosh->offset==0 && bosh->cursor < bosh->header)
    bosh->cursor = bosh->header;
  bosh->line = bosh->offset + bosh->cursor;
  return n;
}

int keydown(int n) {
  if(!n)
    n = bosh->cursormovement;
  if(bosh->line+bosh->cursormovement>=list_items(bosh->buf))
    return 0;
  if(bosh->cursor + n < ULINES)
    bosh->cursor += n;
  else
    bosh->offset += n;
  bosh->line = bosh->offset + bosh->cursor;
  return n;
}
