/*
bosh system stuff (spawning processes, command line arguments too)

$Id: system.c,v 1.37 2008/12/19 16:51:45 alexsisson Exp $

(C) Copyright 2004-2008 Alex Sisson (alexsisson@gmail.com)

This file is part of bosh.

bosh is free software; you can redistribute it and/or modify
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

bosh is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with bosh; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/stat.h>

#include "system.h"
#include "bosh.h"
#include "misc.h"
#include "rc.h"

char *args = NULL; /* all the arguments concatenated into one, for easy passing to children */
char *shfmt = "cat %s | %s -s -- %s";

int bosh_parseargs(int *pargc, char ***pargv) {
  int i,n;
  int argc = *pargc;
  char **argv = *pargv;

  argv[0] = strdup(PACKAGE);

  struct option opts[] = { {"help",0,0,'h'},
                           {"version",0,0,'v'},
                           {"autorefresh",1,0,'c'},
                           {"command",1,0,'c'},
                           {"cursorsize",1,0,'s'},
                           {"cursormovement",1,0,'m'},
                           {"multilineseperator",1,0,'M'},
                           {"preaction",1,0,'p'},
                           {"refresh",1,0,'r'},
                           {"shell",1,0,'S'},
                           {"header",1,0,'H'},
                           {"footer",1,0,'F'},
                           {0,0,0,0} };

#ifdef LOG
  fprintf(bosh_log,"arg: pre-processing:\n");
  for(i=1;i<argc;i++)
    fprintf(bosh_log,"arg: %d: %s\n",i,argv[i]);
#endif

  /* find first non option arg */
  for(i=1;i<argc;i++) {
    if(argv[i][0]!='-') {
      confpath = argv[i];
      argv[i] = strdup("--");
      break;
    }
  }

  if(confpath) {
    conf = strrchr(confpath,'/');
    conf = conf ? conf+1 : confpath;
    conf = strdup(conf);
    rc_read(confpath,bosh,pargc,pargv);
    argc = *pargc;
    argv = *pargv;
#ifdef LOG
  fprintf(bosh_log,"arg: after rc_read:\n");
  for(i=1;i<argc;i++)
    fprintf(bosh_log,"arg: %d: %s\n",i,argv[i]);
#endif
  }

  while(1) {
    n = getopt_long(argc,argv,"hv",opts,0);
    if(n<0)
      break;
    switch(n) {

      case 'a':
        bosh->autorefresh = strtol(optarg,NULL,10);
        break;

      case 'c':
#ifdef LOG
  fprintf(bosh_log,"arg: command: %s\n",optarg);
#endif
        bosh->command = strdup(optarg);
        break;

      case 's':
        bosh->cursorsize = strtol(optarg,NULL,10);
        if(bosh->cursormovement==1)
          bosh->cursormovement = bosh->cursorsize;
        break;

      case 'm':
        bosh->cursormovement = strtol(optarg,NULL,10);
        break;

      case 'M':
        bosh->multilineseperator = *optarg ? strdup(optarg) : NULL;
        break;

      case 'p':
        bosh->preaction = strdup2(optarg,"\n");
        break;

      case 'r':
        bosh->refresh = (*optarg == '1');
        break;

      case 'S':
        shell = strdup(optarg);
        break;

      case 'H':
        bosh->header = strtol(optarg,NULL,10);
        break;

      case 'F':
        bosh->footer = strtol(optarg,NULL,10);
        break;

      case 'h':
        printf("usage:\n\n");
        exit(0);

      case 'v':
        printf("%s %s",PACKAGE,VERSION);
#ifdef DEBUG
        printf(" (DEBUG)");
#endif
#ifdef LOG
        printf(" (LOG)");
#endif
        printf("\n");
        exit(0);

      case '?':
        exit(1);
    }
  }

#ifdef LOG
  fprintf(bosh_log,"arg: post-processing:\n");
  for(i=1;i<argc;i++)
    fprintf(bosh_log,"arg: %d: %s\n",i,argv[i]);
  fprintf(bosh_log,"arg: from optind:\n");
  for(i=optind;i<argc;i++)
    fprintf(bosh_log,"arg: %d: %s\n",i,argv[i]);
#endif

  /* now assemble rest of arguments into big arg string */
  n=6; /* 6 = secret $1 ("bosh"), a space, and a \0 */
  for(i=optind;i<argc;i++)
    n += strlen(argv[i]) + 3; /* +3 for 2 quotes and a space */
  args = malloc(n);
  if(!args)
    return -1;
  strcpy(args,"bosh "); /* secret $1 */
  for(i=optind;i<argc;i++)
    sprintf(args+strlen(args),"'%s' ",argv[i]);

  return 0;
}


int bosh_popen(BOSH *bosh, char *cmd) {

  int fd[2],parent[3],child[3],n,r=-1;
  pid_t pid;

  if(bosh->childpid) {
//    return -1;
    kill(bosh->childpid,3);
    usleep(500);
  }

  /* check that std{in,out,err} are 0,1,2 */
  if(0!=STDIN_FILENO || 1!=STDOUT_FILENO || 2!=STDERR_FILENO)
    return -1;

  /* memset our pipes to invalid fd's */
  memset(parent,-1,sizeof(int)*3);
  memset(child, -1,sizeof(int)*3);

  /* create a pipe for each std stream */
  for(n=0;n<=2;n++) {
    if(pipe(fd))
      goto pipecleanup;
    if(!n) {
      /* stdin */
      parent[n] = fd[1];
      child [n] = fd[0];
    } else {
      /* stdout/err */
      parent[n] = fd[0];
      child [n] = fd[1];
    }
  }

  strcpy(bosh->script,"/tmp/boshXXXXXX");
  close(mkstemp(bosh->script));

  pid = fork();

  if(pid>0) {
    /* parent */

    /* close child's ends of the pipes */
    for(n=0;n<=2;n++)
      close(child[n]);

    /* see if child has died already due to an invalid command */
    usleep(BOSH_PIPE_USLEEP);
    if(pid==waitpid(pid,&n,WNOHANG))
      if(WIFEXITED(n))
        if(WEXITSTATUS(n)==127)
          goto pipecleanup;

    /* input fd is now the child's stdout */
    bosh->fd = parent[1];
    bosh->fderr = parent[2];
    close(parent[0]);

    /* clear the buffer */
    list_reinit(bosh->buf);

    bosh->childpid = pid;
    bosh->exit = -1;
    return 0;

  } else if(pid==0) {
    /* child */
    char *v[4];
    FILE *f;
    char tmpbuf[16];

    /* close parent's ends of the pipes, and dup our ends onto our std streams before closing them */
    for(n=0;n<=2;n++) {
      close(parent[n]);
      dup2(child[n],n);
      close(child[n]);
    }

    if(cmd) {
      /* we're doing a pipe - preserve current buffer to tmp file */
      char *p;
      strcpy(tmpbuf,"/tmp/boshXXXXXX");
      f = fdopen(mkstemp(tmpbuf),"w");
      list_start(bosh->buf,0);
      while(1) {
        p = list_next(bosh->buf);
        if(!p)
          break;
        fprintf(f,"%s\n",p);
      }
      fclose(f);
    }

    /* create a temporary script for the action */
    f = fopen(bosh->script,"w");
    if(!f)
      return -1;

    /* skip secret $1 argument */
    fprintf(f,"shift\n");

    /* set BOSH variables*/
//    fprintf(f,"BOSH='%s'\n",dollarbosh);
    fprintf(f,"BOSHERRFILE=%s\n",dollarbosherrfile);
    fprintf(f,"BOSHCONF=%s\n",conf);
    fprintf(f,"BOSHPID=%d\n",getpid());
    fprintf(f,"BOSHPPID=%d\n",getppid());
    if(dollarboshparam)
      fprintf(f,"BOSHPARAM=%s\n",dollarboshparam);

    /* start function */
    fprintf(f,"function BOSHFUNCTION() {\n");

    /* commands */
    if(cmd) {
      fprintf(f,"cat %s | %s\n",tmpbuf,cmd);
      fprintf(f,"BOSHRETURN=$?\n");
      fprintf(f,"rm -f %s\n",tmpbuf);
    } else {
      fprintf(f,"%s\n",bosh->command);
      fprintf(f,"BOSHRETURN=$?\n");
    }

    /* end function */
    fprintf(f,"}\n");

    /* call function */
    fprintf(f,"BOSHFUNCTION \"$@\"\n");

    /* handle BOSHERR */
    fprintf(f,"[ \"$BOSHERR\" ] && echo $BOSHERR >> $BOSHERRFILE\n");

    /* return */
    fprintf(f,"exit $BOSHRETURN\n");

    /* close */
    fclose(f);
    chmod(bosh->script,0700);

    /* args for exec */
    v[0] = "/bin/sh";
    v[1] = "-c";
    v[2] = malloc(strlen(shfmt)+strlen(bosh->script)+strlen(shell)+strlen(args)+1);
    sprintf(v[2],shfmt,bosh->script,shell,args);
    v[3] = 0;

    /* exec */
    execv(*v,v);

    /* exec failed if we get here - drop through to cleanup. */
    /* or not - this pipecleanup business needs sorting */
    _exit(1);
  }

pipecleanup:
  for(n=0;n<=2;n++) {
    close(parent[n]);
    close(child[n]);
  }

  return r;
}


int bosh_open(BOSH *bosh) {
  return bosh_popen(bosh,NULL);
}

int bosh_pipe(BOSH *bosh, char *cmd) {
  return bosh_popen(bosh,cmd);
}

int bosh_close(BOSH *bosh) {
  return 0;
}
